part of pdf;

/// The [PdfPageLayer] used to create layers in PDF document. Layers refers to sections of content in a PDF document that can be selectively viewed or hidden by document authors or consumers.
class PdfPageLayer implements _IPdfWrapper {
  //Constructor
  /// Initializes a new instance of the [PdfPageLayer] class with specified PDF page.
  PdfPageLayer(PdfPage pdfPage) {
    _initialize(pdfPage, true);
  }

  PdfPageLayer._fromClipPageTemplate(PdfPage pdfPage,
      [bool clipPageTemplates]) {
    _initialize(pdfPage, clipPageTemplates);
  }

  //Fields
  _PdfStream _content;
  PdfPage _page;
  bool _clipPageTemplates;
  //ignore:unused_field
  String _name;
  PdfGraphics _graphics;
  //ignore:unused_field
  PdfColorSpace _colorSpace;
  PdfGraphicsState _graphicsState;
  bool _isEndState = false;
  bool _isSaved = false;

  //Properties
  /// Gets parent page of the layer.
  PdfPage get page => _page;

  /// Gets Graphics context of the layer, used to draw various graphical content on layer.
  PdfGraphics get graphics {
    if (_graphics == null || _isSaved) {
      _initializeGraphics(page);
    }
    return _graphics;
  }

  //Implementation
  void _initialize(PdfPage pdfPage, bool clipPageTemplates) {
    if (pdfPage != null) {
      _page = pdfPage;
    } else {
      throw ArgumentError.value(pdfPage, 'page');
    }
    _clipPageTemplates = clipPageTemplates;
    _content = _PdfStream();
  }

  void _initializeGraphics(PdfPage page) {
    final PdfPage pdfPage = page;
    if (_graphics == null) {
      final Function resources = page._getResources;
      bool isPageHasMediaBox = false;
      if (page._dictionary
          .containsKey(_PdfName(_DictionaryProperties.mediaBox))) {
        isPageHasMediaBox = true;
      }
      final _PdfArray mediaBox = page._dictionary._getValue(
              _DictionaryProperties.mediaBox, _DictionaryProperties.parent)
          as _PdfArray;
      double uppderRightY = 0;
      if (mediaBox != null) {
        uppderRightY = (mediaBox[3] as _PdfNumber).value;
      }
      _graphics = PdfGraphics._(page.size, resources, _content);
      if (isPageHasMediaBox) {
        _graphics._mediaBoxUpperRightBound = uppderRightY;
      }
      if (pdfPage != null) {
        final PdfSectionCollection sectionCollection = pdfPage._section._parent;
        if (sectionCollection != null) {
          _graphics.colorSpace = sectionCollection._document.colorSpace;
          _colorSpace = sectionCollection._document.colorSpace;
        }
      }
      _content._beginSave = _beginSaveContent;
    }
    _graphicsState = _graphics.save();
    _graphics._initializeCoordinates();
    if (_graphics._hasTransparencyBrush) {
      _graphics._setTransparencyGroup(page);
    }
    if (pdfPage != null) {
      final _Rectangle clipRect =
          pdfPage._section._getActualBounds(pdfPage, true);
      if (_clipPageTemplates) {
        if (page._origin.dx >= 0 && page._origin.dy >= 0) {
          _graphics._clipTranslateMarginsWithBounds(clipRect);
        }
      } else {
        final PdfMargins margins = pdfPage._section.pageSettings.margins;
        _graphics._clipTranslateMargins(clipRect.x, clipRect.y, margins.left,
            margins.top, margins.right, margins.bottom);
      }
    }
    _graphics._setLayer(this);
    _isSaved = false;
  }

  void _beginSaveContent(Object sender, _SavePdfPrimitiveArgs args) {
    if (_graphicsState != null) {
      if (_isEndState) {
        graphics._streamWriter._write('EMC\n');
        _isEndState = false;
      }
      graphics.restore(_graphicsState);
      _graphicsState = null;
    }
    _isSaved = true;
  }

  //_IPdfWrapper elements
  @override
  _IPdfPrimitive get _element => _content;
  @override
  //ignore: unused_element
  set _element(_IPdfPrimitive value) {
    _content = value;
  }
}
